from unittest import TestCase

from rapt2.treebrd.errors import RelationReferenceError
from rapt2.treebrd.schema import Schema
from rapt2.treebrd.node import DefinitionNode


class TestSchema(TestCase):
    def test_contains_when_empty(self):
        self.assertFalse(Schema({}).contains("relation"))

    def test_contains_when_false(self):
        self.assertFalse(Schema({"another_relation": []}).contains("relation"))

    def test_contains_when_true(self):
        self.assertTrue(Schema({"relation": []}).contains("relation"))

    def test_to_dict(self):
        expected = {"alpha": ["a1"], "beta": ["b1"]}
        actual = Schema(expected).to_dict()
        self.assertNotEqual(id(expected), id(actual))
        self.assertEqual(expected, actual)

    def test_get_attributes(self):
        raw = {"alpha": ["a1"], "beta": ["b1"]}
        expected = ["a1"]
        actual = Schema(raw).get_attributes("alpha")
        self.assertNotEqual(id(expected), id(raw["alpha"]))
        self.assertEqual(expected, actual)

    def test_add(self):
        schema = Schema({"alpha": ["a1"]})
        schema.add("beta", ["b1"])
        self.assertTrue(schema.contains("beta"))
        self.assertEqual(["b1"], schema.get_attributes("beta"))

    def test_exception_when_name_conflicts(self):
        schema = Schema({"alpha": ["a1"]})
        self.assertRaises(RelationReferenceError, schema.add, "alpha", [])

    def test_definition_node_adds_to_schema(self):
        """Test that DefinitionNode adds its attributes to the schema."""
        schema = Schema({"existing": ["attr1"]})
        
        # Create a definition node
        definition = DefinitionNode("new_relation", ["attr1", "attr2", "attr3"], schema)
        
        # Verify the relation was added to the schema
        self.assertTrue(schema.contains("new_relation"))
        self.assertEqual(["attr1", "attr2", "attr3"], schema.get_attributes("new_relation"))
        
        # Verify the existing relation is still there
        self.assertTrue(schema.contains("existing"))
        self.assertEqual(["attr1"], schema.get_attributes("existing"))

    def test_definition_node_raises_error_on_duplicate_name(self):
        """Test that DefinitionNode raises error when trying to define a relation that already exists."""
        schema = Schema({"existing": ["attr1"]})
        
        # This should raise a RelationReferenceError
        self.assertRaises(RelationReferenceError, DefinitionNode, "existing", ["attr2", "attr3"], schema)



from rapt2.treebrd.condition_node import (BinaryConditionalOperator,
                                          BinaryConditionNode,
                                          IdentityConditionNode)
from rapt2.treebrd.errors import AttributeReferenceError, InputError
from rapt2.treebrd.node import (BinaryNode, CrossJoinNode, DifferenceNode,
                                FullOuterJoinNode, IntersectNode,
                                LeftOuterJoinNode, Operator,
                                RightOuterJoinNode, ThetaJoinNode, UnionNode)
from tests.treebrd.test_node import NodeTestCase


class BinaryTestCase(NodeTestCase):
    pass


class TestBinaryNode(BinaryTestCase):
    def test_binary_children(self):
        node = BinaryNode(Operator.union, self.beta, self.gamma)
        self.assertEqual(self.beta, node.left)
        self.assertEqual(self.gamma, node.right)


class TestBinaryNodeEquality(BinaryTestCase):
    def test_equality_when_identical(self):
        node = BinaryNode(Operator.cross_join, self.alpha, self.beta, "borg")
        same = node
        self.assertTrue(node == same)

    def test_equality_when_same_operator_and_name(self):
        node = BinaryNode(Operator.cross_join, self.alpha, self.beta, "borg")
        twin = BinaryNode(Operator.cross_join, self.alpha, self.beta, "borg")
        self.assertTrue(node == twin)

    def test_non_equality_when_different_operator_and_name(self):
        node = BinaryNode(Operator.natural_join, self.alpha, self.beta, "borg")
        other = BinaryNode(Operator.cross_join, self.alpha, self.beta, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_different_operator(self):
        node = BinaryNode(Operator.natural_join, self.alpha, self.beta, "borg")
        other = BinaryNode(Operator.cross_join, self.alpha, self.beta, "borg")
        self.assertTrue(node != other)

    def test_non_equality_when_different_name(self):
        node = BinaryNode(Operator.cross_join, self.alpha, self.beta, "borg")
        other = BinaryNode(Operator.cross_join, self.alpha, self.beta, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_same_operator_and_one_no_name(self):
        node = BinaryNode(Operator.cross_join, self.alpha, self.beta, "borg")
        other = BinaryNode(Operator.cross_join, self.alpha, self.beta)
        self.assertTrue(node != other)


class TestUnionNode(BinaryTestCase):
    def test_operator_on_init(self):
        node = UnionNode(self.beta, self.beta)
        self.assertEqual(Operator.union, node.operator)

    def test_children_on_init(self):
        node = UnionNode(self.twin, self.twin_prime)
        self.assertEqual(self.twin, node.left)
        self.assertEqual(self.twin_prime, node.right)

    def test_attributes(self):
        node = UnionNode(self.beta, self.beta)
        expected = get_attributes("beta", self.schema)
        self.assertEqual(expected, node.attributes.to_list())

    def test_mismatch(self):
        self.assertRaises(InputError, UnionNode, self.beta, self.gamma)


class TestDifferenceNode(BinaryTestCase):
    def test_operator_on_init(self):
        node = DifferenceNode(self.beta, self.beta)
        self.assertEqual(Operator.difference, node.operator)

    def test_name_on_init(self):
        node = DifferenceNode(self.beta, self.beta)
        self.assertIsNone(node.name)

    def test_children_on_init(self):
        node = DifferenceNode(self.twin, self.twin_prime)
        self.assertEqual(self.twin, node.left)
        self.assertEqual(self.twin_prime, node.right)

    def test_attributes(self):
        node = DifferenceNode(self.beta, self.beta)
        expected = get_attributes("beta", self.schema)
        self.assertEqual(expected, node.attributes.to_list())

    def test_mismatch(self):
        self.assertRaises(InputError, DifferenceNode, self.beta, self.gamma)


class TestIntersectionNode(BinaryTestCase):
    def test_operator_on_init(self):
        node = IntersectNode(self.beta, self.beta)
        self.assertEqual(Operator.intersect, node.operator)

    def test_children_on_init(self):
        node = IntersectNode(self.twin, self.twin_prime)
        self.assertEqual(self.twin, node.left)
        self.assertEqual(self.twin_prime, node.right)

    def test_attributes(self):
        node = IntersectNode(self.beta, self.beta)
        expected = get_attributes("beta", self.schema)
        self.assertEqual(expected, node.attributes.to_list())

    def test_mismatch(self):
        self.assertRaises(InputError, IntersectNode, self.beta, self.gamma)


class TestJoinNode(BinaryTestCase):
    def test_operator_on_init(self):
        node = CrossJoinNode(self.alpha, self.beta)
        self.assertEqual(Operator.cross_join, node.operator)

    def test_children_on_init(self):
        node = CrossJoinNode(self.beta, self.gamma)
        self.assertEqual(self.beta, node.left)
        self.assertEqual(self.gamma, node.right)

    def test_attributes_on_init(self):
        node = CrossJoinNode(self.beta, self.gamma)
        expected = get_attributes(
            "beta", self.schema, use_prefix=True
        ) + get_attributes("gamma", self.schema, use_prefix=True)
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_left_child_is_join(self):
        left = CrossJoinNode(self.beta, self.gamma)
        node = CrossJoinNode(left, self.alpha)
        expected = (
            get_attributes("beta", self.schema, use_prefix=True)
            + get_attributes("gamma", self.schema, use_prefix=True)
            + get_attributes("alpha", self.schema, use_prefix=True)
        )
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_right_child_is_join(self):
        right = CrossJoinNode(self.gamma, self.alpha)
        node = CrossJoinNode(self.beta, right)
        expected = (
            get_attributes("beta", self.schema, use_prefix=True)
            + get_attributes("gamma", self.schema, use_prefix=True)
            + get_attributes("alpha", self.schema, use_prefix=True)
        )
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_both_children_are_join(self):
        left = CrossJoinNode(self.alpha, self.beta)
        right = CrossJoinNode(self.gamma, self.twin)
        node = CrossJoinNode(left, right)
        expected = (
            get_attributes("alpha", self.schema, use_prefix=True)
            + get_attributes("beta", self.schema, use_prefix=True)
            + get_attributes("gamma", self.schema, use_prefix=True)
            + get_attributes("twin", self.schema, use_prefix=True)
        )
        self.assertEqual(expected, node.attributes.to_list())


class TestCrossNode(BinaryTestCase):
    def test_operator_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = ThetaJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(Operator.theta_join, node.operator)

    def test_children_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = ThetaJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(self.alpha, node.left)
        self.assertEqual(self.beta, node.right)

    def test_attributes_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = ThetaJoinNode(self.alpha, self.beta, condition)
        expected = get_attributes(
            "alpha", self.schema, use_prefix=True
        ) + get_attributes("beta", self.schema, use_prefix=True)
        self.assertEqual(expected, node.attributes.to_list())

    def test_condition_when_init_has_condition(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        actual = ThetaJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = ThetaJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.GREATER_THAN,
            IdentityConditionNode("alpha.a1"),
            IdentityConditionNode("41"),
        )
        actual = ThetaJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("alpha.a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("beta.b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = ThetaJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_exception_when_first_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            ThetaJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_second_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            ThetaJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a2"),
            ),
        )

    def test_exception_when_both_attributes_in_condition_are_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            ThetaJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a3"),
            ),
        )

    def test_exception_when_prefix_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            ThetaJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("beta.a1"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_ambiguous_attributes(self):
        self.assertRaises(
            AttributeReferenceError,
            ThetaJoinNode,
            self.alpha,
            self.ambiguous,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("d1"),
                IdentityConditionNode("42"),
            ),
        )


class TestFullOuterJoinNode(BinaryTestCase):
    def test_operator_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = FullOuterJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(Operator.full_outer_join, node.operator)

    def test_children_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = FullOuterJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(self.alpha, node.left)
        self.assertEqual(self.beta, node.right)

    def test_attributes_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = FullOuterJoinNode(self.alpha, self.beta, condition)
        expected = get_attributes(
            "alpha", self.schema, use_prefix=True
        ) + get_attributes("beta", self.schema, use_prefix=True)
        self.assertEqual(expected, node.attributes.to_list())

    def test_condition_when_init_has_condition(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        actual = FullOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = FullOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.GREATER_THAN,
            IdentityConditionNode("alpha.a1"),
            IdentityConditionNode("41"),
        )
        actual = FullOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("alpha.a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("beta.b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = FullOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_exception_when_first_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            FullOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_second_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            FullOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a2"),
            ),
        )

    def test_exception_when_both_attributes_in_condition_are_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            FullOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a3"),
            ),
        )

    def test_exception_when_prefix_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            FullOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("beta.a1"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_ambiguous_attributes(self):
        self.assertRaises(
            AttributeReferenceError,
            FullOuterJoinNode,
            self.alpha,
            self.ambiguous,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("d1"),
                IdentityConditionNode("42"),
            ),
        )


class TestLeftOuterJoinNode(BinaryTestCase):
    def test_operator_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = LeftOuterJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(Operator.left_outer_join, node.operator)

    def test_children_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = LeftOuterJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(self.alpha, node.left)
        self.assertEqual(self.beta, node.right)

    def test_attributes_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = LeftOuterJoinNode(self.alpha, self.beta, condition)
        expected = get_attributes(
            "alpha", self.schema, use_prefix=True
        ) + get_attributes("beta", self.schema, use_prefix=True)
        self.assertEqual(expected, node.attributes.to_list())

    def test_condition_when_init_has_condition(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        actual = LeftOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = LeftOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.GREATER_THAN,
            IdentityConditionNode("alpha.a1"),
            IdentityConditionNode("41"),
        )
        actual = LeftOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("alpha.a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("beta.b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = LeftOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_exception_when_first_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            LeftOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_second_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            LeftOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a2"),
            ),
        )

    def test_exception_when_both_attributes_in_condition_are_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            LeftOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a3"),
            ),
        )

    def test_exception_when_prefix_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            LeftOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("beta.a1"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_ambiguous_attributes(self):
        self.assertRaises(
            AttributeReferenceError,
            LeftOuterJoinNode,
            self.alpha,
            self.ambiguous,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("d1"),
                IdentityConditionNode("42"),
            ),
        )


class TestRightOuterJoinNode(BinaryTestCase):
    def test_operator_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = RightOuterJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(Operator.right_outer_join, node.operator)

    def test_children_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = RightOuterJoinNode(self.alpha, self.beta, condition)
        self.assertEqual(self.alpha, node.left)
        self.assertEqual(self.beta, node.right)

    def test_attributes_on_init(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        node = RightOuterJoinNode(self.alpha, self.beta, condition)
        expected = get_attributes(
            "alpha", self.schema, use_prefix=True
        ) + get_attributes("beta", self.schema, use_prefix=True)
        self.assertEqual(expected, node.attributes.to_list())

    def test_condition_when_init_has_condition(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        actual = RightOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = RightOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.GREATER_THAN,
            IdentityConditionNode("alpha.a1"),
            IdentityConditionNode("41"),
        )
        actual = RightOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("alpha.a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("beta.b1"),
                IdentityConditionNode("43"),
            ),
        )
        actual = RightOuterJoinNode(self.alpha, self.beta, condition).conditions
        self.assertEqual(condition, actual)

    def test_exception_when_first_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            RightOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_second_attribute_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            RightOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a2"),
            ),
        )

    def test_exception_when_both_attributes_in_condition_are_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            RightOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("a2"),
                IdentityConditionNode("a3"),
            ),
        )

    def test_exception_when_prefix_in_condition_is_wrong(self):
        self.assertRaises(
            AttributeReferenceError,
            RightOuterJoinNode,
            self.alpha,
            self.beta,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("beta.a1"),
                IdentityConditionNode("42"),
            ),
        )

    def test_exception_when_ambiguous_attributes(self):
        self.assertRaises(
            AttributeReferenceError,
            RightOuterJoinNode,
            self.alpha,
            self.ambiguous,
            BinaryConditionNode(
                BinaryConditionalOperator.EQUAL,
                IdentityConditionNode("d1"),
                IdentityConditionNode("42"),
            ),
        )


def get_attributes(name, schema, use_prefix=False):
    def prefixed(attribute):
        if use_prefix:
            return "{name}.{attr}".format(name=name, attr=attribute)
        else:
            return attribute

    return [prefixed(attribute) for attribute in schema.get_attributes(name)]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto

from .grammars.condition_grammar import ConditionGrammar
from .grammars.syntax import Syntax


class UnaryConditionalOperator(Enum):
    NOT = auto()
    DEFINED = auto()

    @classmethod
    def from_syntax(cls, syntax: Syntax, instring: str) -> "UnaryConditionalOperator":
        match instring:
            case syntax.not_op:
                return cls.NOT
            case syntax.defined_op:
                return cls.DEFINED
            case _:
                raise ValueError


class BinaryConditionalOperator(Enum):
    AND = auto()
    OR = auto()
    EQUAL = auto()
    NOT_EQUAL = auto()
    LESS_THAN = auto()
    LESS_THAN_EQUAL = auto()
    GREATER_THAN = auto()
    GREATER_THAN_EQUAL = auto()

    @classmethod
    def from_syntax(cls, syntax: Syntax, instring: str) -> "BinaryConditionalOperator":
        match instring:
            case syntax.and_op:
                return cls.AND
            case syntax.or_op:
                return cls.OR
            case syntax.equal_op:
                return cls.EQUAL
            case syntax.not_equal_op | syntax.not_equal_alt_op:
                return cls.NOT_EQUAL
            case syntax.less_than_op:
                return cls.LESS_THAN
            case syntax.less_than_equal_op:
                return cls.LESS_THAN_EQUAL
            case syntax.greater_than_op:
                return cls.GREATER_THAN
            case syntax.greater_than_equal_op:
                return cls.GREATER_THAN_EQUAL
            case _:
                raise ValueError


@dataclass(frozen=True)
class ConditionNode(ABC):
    @abstractmethod
    def attribute_references(self) -> list[str]:
        """
        Return a list of attribute references in this condition node.
        """
        raise NotImplementedError()


@dataclass(frozen=True)
class IdentityConditionNode(ConditionNode):
    ident: str

    def attribute_references(self) -> list[str]:
        try:
            ConditionGrammar().attribute_reference.parse_string(self.ident)
            return [self.ident]
        except Exception:
            return []


@dataclass(frozen=True)
class UnaryConditionNode(ConditionNode):
    op: UnaryConditionalOperator
    child: ConditionNode

    def attribute_references(self) -> list[str]:
        return self.child.attribute_references()


@dataclass(frozen=True)
class BinaryConditionNode(ConditionNode):
    op: BinaryConditionalOperator
    left: ConditionNode
    right: ConditionNode

    def attribute_references(self) -> list[str]:
        return list(
            self.left.attribute_references() + self.right.attribute_references()
        )



from unittest import TestCase

from rapt2.treebrd.attributes import AttributeList
from rapt2.treebrd.errors import RelationReferenceError
from rapt2.treebrd.node import Node, Operator, RelationNode
from rapt2.treebrd.schema import Schema


class NodeTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        # logging.basicConfig(level=logging.NOTSET)
        pass

    def setUp(self):
        self.schema = Schema(
            {
                "alpha": ["a1"],
                "beta": ["b1", "b2"],
                "gamma": ["c1", "c2", "c3"],
                "twin": ["t1", "t2", "t3"],
                "twin_prime": ["t1", "t2", "t3"],
                "ambiguous": ["d1", "d1"],
            }
        )
        self.alpha = RelationNode("alpha", self.schema)
        self.beta = RelationNode("beta", self.schema)
        self.gamma = RelationNode("gamma", self.schema)
        self.twin = RelationNode("twin", self.schema)
        self.twin_prime = RelationNode("twin_prime", self.schema)
        self.ambiguous = RelationNode("ambiguous", self.schema)


class TestNode(TestCase):
    def test_operator_from_init(self):
        expected = Operator.relation
        actual = Node(Operator.relation, None).operator
        self.assertEqual(expected, actual)

    def test_name_when_init_has_none(self):
        node = Node(Operator.relation, None)
        self.assertIsNone(node.name)

    def test_name_when_init_has_name(self):
        expected = "alpha"
        actual = Node(Operator.relation, "alpha").name
        self.assertEqual(expected, actual)

    def test_attributes_from_init_when_none(self):
        actual = Node(Operator.relation, None).attributes
        self.assertIsNone(actual)


class TestNodeEquality(TestCase):
    def test_equality_when_identical(self):
        node = Node(Operator.relation, "borg")
        same = node
        self.assertTrue(node == same)

    def test_equality_when_same_operator_and_name(self):
        node = Node(Operator.relation, "borg")
        twin = Node(Operator.relation, "borg")
        self.assertTrue(node == twin)

    def test_non_equality_when_different_operator_and_name(self):
        node = Node(Operator.relation, "borg")
        other = Node(Operator.project, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_different_operator(self):
        node = Node(Operator.relation, "borg")
        other = Node(Operator.project, "borg")
        self.assertTrue(node != other)

    def test_non_equality_when_different_name(self):
        node = Node(Operator.relation, "borg")
        other = Node(Operator.relation, "other")
        self.assertTrue(node != other)


class TestRelationNode(NodeTestCase):
    def test_operator_from_init(self):
        expected = Operator.relation
        actual = RelationNode("alpha", self.schema).operator
        self.assertEqual(expected, actual)

    def test_exception_raised_when_name_not_in_schema(self):
        self.assertRaises(RelationReferenceError, RelationNode, "unknown", self.schema)

    def test_attributes_when_name_is_in_schema(self):
        expected = AttributeList(self.schema.get_attributes("alpha"), "alpha").to_list()
        node = RelationNode("alpha", self.schema)
        self.assertEqual(expected, node.attributes.to_list())



import copy
from enum import Enum, auto

from .attributes import AttributeList
from .condition_node import ConditionNode
from .errors import InputError, RelationReferenceError


class Node:
    """
    A Node from a relational algebra parse tree.

    A node represents three intermingled concepts:
        - A relational algebra operator and the parameters requires
        - Together with its children, a single relational algebra expression
        - A method of mutating a collection of tuples
    """

    name: str | None

    def __init__(self, operator, name=None):
        """
        Construct a node.
        :param name: The name of the relation.
        :param operator: The operator used to create this relation. One of
        the node.Operator enum.
        """
        self.operator = operator
        self.name = name
        self.attributes = None

    def __eq__(self, other):
        """
        Return true if other is a node with the same operator, name and
        attributes. Else return false.
        :param other: A Node.
        :return: True if other is equivalent to this node.
        """
        if type(self) is not type(other):
            return False
        if self.name != other.name:
            return False
        if self.operator != other.operator:
            return False
        if self.attributes != other.attributes:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class RelationNode(Node):
    """
    A relation.
    """

    def __init__(self, name: str, schema):
        super().__init__(Operator.relation, name)
        self.attributes = AttributeList(schema.get_attributes(name), name)

    def post_order(self):
        return [self]


class DefinitionNode(Node):
    """
    A relation definition.
    """

    def __init__(self, name: str, attributes: list, schema):
        super().__init__(Operator.definition, name)
        self.attributes = AttributeList(attributes, name)
        schema.add(name, self.attributes.names)

    def post_order(self):
        return [self]


class UnaryNode(Node):
    """
    A Node with one child.
    """

    def __init__(self, operator, child, name: str | None = None):
        super().__init__(operator, name)
        self.child = child

        if not name:
            self.name = self.child.name

        self.attributes = copy.copy(child.attributes)

    def __eq__(self, other):
        return super().__eq__(other) and self.child.__eq__(other.child)

    def post_order(self):
        return self.child.post_order() + [self]


class SelectNode(UnaryNode):
    """
    A relation that results from the relation algebra select operator.
    """

    def __init__(self, child, conditions: ConditionNode):
        super().__init__(Operator.select, child)
        self.attributes.validate(conditions.attribute_references())
        self.conditions = conditions

    def __eq__(self, other):
        return self.conditions == other.conditions and super().__eq__(other)


class ProjectNode(UnaryNode):
    """
    A relation that results from the relation algebra project operator.
    """

    def __init__(self, child, attributes):
        super().__init__(Operator.project, child=child)
        self.attributes.trim(attributes)


class RenameNode(UnaryNode):
    """
    A relation that results from the relation algebra rename operator.
    """

    def __init__(self, child, name, attributes, schema):
        """
        Construct a RenameNode.
        :param name: The new name for the relation.
        :param attributes: A list of new names for every attribute or an empty
        list.
        :param child: The child of this Node.
        """
        super().__init__(Operator.rename, child, name)
        if schema.contains(name):
            raise RelationReferenceError(
                "Relation '{name}' already exists.".format(name=name)
            )
        self.attributes.rename(attributes, self.name)


class AssignNode(UnaryNode):
    """
    A relation that results from the relation algebra assign operator.
    """

    def __init__(self, child, name, attributes, schema):
        """
        Construct an AssignNode.
        :param name: The new name for the relation.
        :param attributes: A list of new names for every attribute or an empty
        list.
        :param child: The child of this Node.
        """
        if not name:
            raise InputError("Name is required for assignment.")

        if attributes and len(attributes) != len(child.attributes):
            raise InputError("Assignment requires naming all attributes.")

        super().__init__(Operator.assign, child, name)
        self.attributes.rename(attributes, name)
        schema.add(name, self.attributes.names)


class BinaryNode(Node):
    """
    A Node with two children, a left and a right.
    """

    def __init__(self, operator, left, right, name=None):
        super().__init__(operator, name)
        self.left = left
        self.right = right

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and self.left.__eq__(other.left)
            and self.right.__eq__(other.right)
        )

    def post_order(self):
        return self.left.post_order() + self.right.post_order() + [self]


class JoinNode(BinaryNode):
    """
    A relation that results from the relation algebra cross join operator.
    """

    def __init__(self, operator, left, right):
        if left.name and right.name and left.name == right.name:
            raise RelationReferenceError("Ambiguous relation reference.")
        super().__init__(operator, left, right, None)
        self.attributes = AttributeList.merge(left.attributes, right.attributes)


class CrossJoinNode(JoinNode):
    """
    A relation that results from the relation algebra cross join operator.
    """

    def __init__(self, left, right):
        super().__init__(Operator.cross_join, left, right)


class NaturalJoinNode(JoinNode):
    """
    A relation that results from the relation algebra natural join operator.
    """

    def __init__(self, left, right):
        super().__init__(Operator.natural_join, left, right)
        left_attributes = [attribute.prefixed for attribute in self.left.attributes]
        left_names = self.left.attributes.names
        right_attributes = [
            attribute.prefixed
            for attribute in self.right.attributes
            if attribute.name not in left_names
        ]
        self.attributes.trim(left_attributes + right_attributes)


class ThetaJoinNode(JoinNode):
    """
    A relation that results from the relation algebra theta join operator.
    """

    def __init__(self, left, right, conditions: ConditionNode):
        super().__init__(Operator.theta_join, left, right)
        self.attributes.validate(conditions.attribute_references())
        self.conditions = conditions

    def __eq__(self, other):
        return self.conditions == other.conditions and super().__eq__(other)


class FullOuterJoinNode(JoinNode):
    """
    A relation that results from the relation algebra full outer join operator.
    """

    def __init__(self, left, right, conditions: ConditionNode):
        super().__init__(Operator.full_outer_join, left, right)
        self.attributes.validate(conditions.attribute_references())
        self.conditions = conditions

    def __eq__(self, other):
        return self.conditions == other.conditions and super().__eq__(other)


class LeftOuterJoinNode(JoinNode):
    """
    A relation that results from the relation algebra left outer join operator.
    """

    def __init__(self, left, right, conditions: ConditionNode):
        super().__init__(Operator.left_outer_join, left, right)
        self.attributes.validate(conditions.attribute_references())
        self.conditions = conditions

    def __eq__(self, other):
        return self.conditions == other.conditions and super().__eq__(other)


class RightOuterJoinNode(JoinNode):
    """
    A relation that results from the relation algebra right outer join operator.
    """

    def __init__(self, left, right, conditions: ConditionNode):
        super().__init__(Operator.right_outer_join, left, right)
        self.attributes.validate(conditions.attribute_references())
        self.conditions = conditions

    def __eq__(self, other):
        return self.conditions == other.conditions and super().__eq__(other)


class SetOperatorNode(BinaryNode):
    """
    An abstract class for binary nodes with set operators.
    """

    def __init__(self, operator, left, right):
        super().__init__(operator, left, right, None)

        names = left.attributes.names
        if not names == right.attributes.names:
            raise InputError("Set operations require identical relation schemas.")

        self.attributes = AttributeList(names, None)


class UnionNode(SetOperatorNode):
    """
    A relation that results from the relation algebra union operator.
    """

    def __init__(self, left, right):
        super().__init__(Operator.union, left, right)


class DifferenceNode(SetOperatorNode):
    """
    A relation that results from the relation algebra difference operator.
    """

    def __init__(self, left, right):
        super().__init__(Operator.difference, left, right)


class IntersectNode(SetOperatorNode):
    """
    A relation that results from the relation algebra intersect operator.
    """

    def __init__(self, left, right):
        super().__init__(Operator.intersect, left, right)


class DependencyNode(Node):
    """
    Base class for dependency nodes.
    """

    def __init__(self, operator, relation_name, attributes):
        super().__init__(operator)
        self.relation_name = relation_name
        self.attributes = attributes

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and self.relation_name == other.relation_name
            and self.attributes == other.attributes
        )

    def post_order(self):
        return [self]


class PrimaryKeyNode(DependencyNode):
    """
    A node representing a primary key dependency.
    """

    def __init__(self, relation_name, attributes):
        super().__init__(Operator.primary_key, relation_name, attributes)


class MultivaluedDependencyNode(DependencyNode):
    """
    A node representing a multivalued dependency.
    """

    child: RelationNode | SelectNode

    def __init__(self, relation_name, attributes, relation_node):
        super().__init__(Operator.multivalued_dependency, relation_name, attributes)
        self.child = relation_node

    def __eq__(self, other):
        return super().__eq__(other) and self.child == other.child


class FunctionalDependencyNode(DependencyNode):
    """
    A node representing a functional dependency.
    """

    child: RelationNode | SelectNode

    def __init__(self, relation_name, attributes, relation_node):
        super().__init__(Operator.functional_dependency, relation_name, attributes)
        self.child = relation_node

    def __eq__(self, other):
        return super().__eq__(other) and self.child == other.child


class InclusionEquivalenceNode(DependencyNode):
    """
    A node representing an inclusion equivalence dependency.
    """

    left_child: RelationNode | SelectNode
    right_child: RelationNode | SelectNode

    def __init__(self, relation_names, attributes, left_child, right_child):
        super().__init__(Operator.inclusion_equivalence, relation_names, attributes)
        self.relation_names = relation_names
        self.left_child = left_child
        self.right_child = right_child

    def __eq__(self, other):
        return (super().__eq__(other) and 
                self.relation_names == other.relation_names and
                self.left_child == other.left_child and
                self.right_child == other.right_child)


class InclusionSubsumptionNode(DependencyNode):
    """
    A node representing an inclusion subsumption dependency.
    """

    left_child: RelationNode | SelectNode
    right_child: RelationNode | SelectNode

    def __init__(self, relation_names, attributes, left_child, right_child):
        super().__init__(Operator.inclusion_subsumption, relation_names, attributes)
        self.relation_names = relation_names
        self.left_child = left_child
        self.right_child = right_child

    def __eq__(self, other):
        return (super().__eq__(other) and 
                self.relation_names == other.relation_names and
                self.left_child == other.left_child and
                self.right_child == other.right_child)


class Operator(Enum):
    """
    A type of relational algebra operation.
    """

    # Basic operations
    relation = auto()
    assign = auto()
    definition = auto()
    project = auto()
    rename = auto()
    select = auto()

    # Join operations.
    cross_join = auto()
    natural_join = auto()
    theta_join = auto()
    full_outer_join = auto()
    left_outer_join = auto()
    right_outer_join = auto()

    # Set operations.
    difference = auto()
    union = auto()
    intersect = auto()

    # Dependency operations.
    primary_key = auto()
    multivalued_dependency = auto()
    functional_dependency = auto()
    inclusion_equivalence = auto()
    inclusion_subsumption = auto()



from typing import Generic, TypeVar

from pyparsing import (Combine, Optional, ParserElement, Regex, Word,
                       alphanums, alphas, pyparsing_common, quotedString,
                       removeQuotes)

from .syntax import BaseSyntax

ParserElement.enablePackrat()

TSyntax = TypeVar("TSyntax", bound=BaseSyntax)


class ProtoGrammar(Generic[TSyntax]):
    """
    A grammar with fundamental rules for characters, strings, and
    numbers.

    The rules are annotated with their BNF equivalents. For a complete
    specification refer to the associated grammar file.
    """

    syntax: TSyntax

    def __init__(self, syntax: TSyntax = BaseSyntax()):
        """
        Initializes a ProtoGrammar. Uses the default syntax if none
        is provided.

        :param syntax: a syntax for this grammar.
        """
        self.syntax = syntax

    def parse(self, instring):
        """
        Defined by descendants.
        :param instring: A string to parse.
        """
        raise NotImplementedError

    @property
    def character(self):
        """
        character ::= letter | digit | "_"
        """
        return alphanums + "_"

    @property
    def number(self):
        """
        number ::= float | integer | natural_number
        """
        return Regex(r"[-+]?[0-9]*\.?[0-9]+")

    @property
    def string_literal(self):
        """
        string_literal ::= "'" string "'" | "\"" string "\""

        Any successful match is converted to a single quoted string to simplify
        post-parsed operations.
        """
        return quotedString.setParseAction(
            lambda s, loc, t: "'{string}'".format(string=removeQuotes(s, loc, t))
        )

    @property
    def identifier(self):
        """
        identifier ::= letter | letter string
        """
        return Word(alphas, self.character).setParseAction(
            pyparsing_common.downcaseTokens
        )

    @property
    def relation_name(self):
        """
        relation_name ::= identifier
        """
        return self.identifier

    @property
    def attribute_name(self):
        """
        attribute_name ::= identifier
        """
        return self.identifier

    @property
    def attribute_reference(self):
        """
        attribute_reference ::= relation_name "." attribute_name |
        attribute_name
        """
        return Combine((Optional(self.relation_name + ".") + self.attribute_name))



import functools
from unittest import TestCase

from rapt2.treebrd.condition_node import (BinaryConditionalOperator,
                                          BinaryConditionNode,
                                          IdentityConditionNode)
from rapt2.treebrd.errors import RelationReferenceError
from rapt2.treebrd.grammars import DependencyGrammar
from rapt2.treebrd.node import (CrossJoinNode, NaturalJoinNode, ProjectNode,
                                RelationNode, ThetaJoinNode)
from rapt2.treebrd.schema import Schema
from rapt2.treebrd.treebrd import TreeBRD


class TreeBRDTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.schema = None

    @classmethod
    def create_build_function(cls, schema):
        builder = TreeBRD(DependencyGrammar())
        return functools.partial(builder.build, schema=schema)


class TestRelation(TreeBRDTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.definition = {
            "letters": ["position", "value"],
            "numbers": ["value", "prime"],
        }
        cls.schema = Schema(cls.definition)
        cls.build = staticmethod(cls.create_build_function(cls.definition))

    def test_build_when_instring_is_single_relation(self):
        forest = self.build("letters;")
        self.assertEqual(1, len(forest))
        expected = RelationNode("letters", self.schema)
        self.assertEqual(expected, forest[0])

    def test_build_when_instring_has_multiple_relations(self):
        instring = "numbers; letters;"
        forest = self.build(instring)
        self.assertEqual(2, len(forest))
        first = RelationNode("numbers", self.schema)
        second = RelationNode("letters", self.schema)
        self.assertEqual(first, forest[0])
        self.assertEqual(second, forest[1])


class TestProject(TreeBRDTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.definition = {
            "magic_wand": [
                "owner",
                "manufacturer",
                "wood",
                "core",
                "length",
                "rigidity",
            ]
        }
        cls.schema = Schema(cls.definition)
        cls.build = staticmethod(cls.create_build_function(cls.definition))

    def test_project_with_single_attr(self):
        instring = r"\project_{owner} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, ["owner"])
        self.assertEqual(expected, forest[0])

    def test_project_with_two_attr(self):
        instring = r"\project_{owner, wood} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, ["owner", "wood"])
        self.assertEqual(expected, forest[0])

    def test_project_with_all_but_one_attr(self):
        attr = self.schema.get_attributes("magic_wand")
        attr.remove("rigidity")
        instring = r"\project_{" + ", ".join(attr) + r"} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, attr)
        self.assertEqual(expected, forest[0])

    def test_project_with_all_attr(self):
        attr = self.schema.get_attributes("magic_wand")
        instring = r"\project_{" + ", ".join(attr) + r"} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, attr)
        self.assertEqual(expected, forest[0])

    def test_project_with_all_attr_shuffled(self):
        attr = self.schema.get_attributes("magic_wand")
        attr.sort()
        instring = r"\project_{" + ", ".join(attr) + r"} magic_wand;"
        forest = self.build(instring)
        child = RelationNode("magic_wand", self.schema)
        expected = ProjectNode(child, attr)
        self.assertEqual(expected, forest[0])


class JoinTestCase(TreeBRDTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.definition = {
            "alpha": ["a1"],
            "beta": ["b1", "b2"],
            "gamma": ["c1", "c2", "c3"],
        }
        cls.schema = Schema(cls.definition)
        cls.build = staticmethod(cls.create_build_function(cls.definition))


class TestJoins(JoinTestCase):
    def test_join_with_natural_join(self):
        instring = "alpha \\join beta \\natural_join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = CrossJoinNode(left, middle)
        expected = NaturalJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_natural_join_with_join(self):
        instring = "alpha \\natural_join beta \\join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = NaturalJoinNode(left, middle)
        expected = CrossJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_join_with_theta_join(self):
        instring = "alpha \\join beta \\theta_join_{a1 = c1} gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = CrossJoinNode(left, middle)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("c1"),
        )
        expected = ThetaJoinNode(intermediate, right, condition)
        self.assertEqual(expected, forest[0])

    def test_theta_join_with_join(self):
        instring = "alpha \\theta_join_{a1 = b1} beta \\join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        intermediate = ThetaJoinNode(left, middle, condition)
        expected = CrossJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])


class TestCrossJoin(JoinTestCase):
    def test_join_two_separate_relations(self):
        instring = "alpha \\join beta;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        right = RelationNode("beta", self.schema)
        expected = CrossJoinNode(left, right)
        self.assertEqual(expected, forest[0])

    def test_join_three_separate_relations(self):
        instring = "alpha \\join beta \\join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = CrossJoinNode(left, middle)
        expected = CrossJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_exception_when_join_two_identical_relations(self):
        left = RelationNode("alpha", self.schema)
        right = RelationNode("alpha", self.schema)
        self.assertRaises(RelationReferenceError, CrossJoinNode, left, right)


class TestThetaJoin(JoinTestCase):
    def test_join_two_separate_relations(self):
        instring = "alpha \\theta_join_{a1 = b1} beta;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        right = RelationNode("beta", self.schema)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        expected = ThetaJoinNode(left, right, condition)
        self.assertEqual(expected, forest[0])

    def test_join_three_separate_relations(self):
        instring = "alpha \\theta_join_{a1 = b1} beta \\theta_join_{a1 = b1} gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("b1"),
        )
        intermediate = ThetaJoinNode(left, middle, condition)
        expected = ThetaJoinNode(intermediate, right, condition)
        self.assertEqual(expected, forest[0])

    def test_exception_when_join_two_identical_relations(self):
        left = RelationNode("alpha", self.schema)
        right = RelationNode("alpha", self.schema)
        self.assertRaises(RelationReferenceError, ThetaJoinNode, left, right, "a1=5")


class TestNaturalJoin(JoinTestCase):
    def test_join_two_separate_relations(self):
        instring = "alpha \\natural_join beta;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        right = RelationNode("beta", self.schema)
        expected = NaturalJoinNode(left, right)
        self.assertEqual(expected, forest[0])

    def test_join_three_separate_relations(self):
        instring = "alpha \\natural_join beta \\natural_join gamma;"
        forest = self.build(instring)
        left = RelationNode("alpha", self.schema)
        middle = RelationNode("beta", self.schema)
        right = RelationNode("gamma", self.schema)
        intermediate = NaturalJoinNode(left, middle)
        expected = NaturalJoinNode(intermediate, right)
        self.assertEqual(expected, forest[0])

    def test_exception_when_join_two_identical_relations(self):
        left = RelationNode("alpha", self.schema)
        right = RelationNode("alpha", self.schema)
        self.assertRaises(RelationReferenceError, NaturalJoinNode, left, right)



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestProject(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_single_attribute(self):
        expression = r"\project_{species} astronauts;"
        expected = [["\\project", ["species"], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_two_attributes(self):
        expression = r"\project_{name, age} astronauts;"
        expected = [["\\project", ["name", "age"], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_multiple_attributes(self):
        expression = r"\project_{name, age, colour, sex, alive} astronauts;"
        expected = [
            ["\\project", ["name", "age", "colour", "sex", "alive"], ["astronauts"]]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_recursive(self):
        expression = r"\project_{species} \project_{age, species} astronauts;"
        expected = [
            [
                "\\project",
                ["species"],
                ["\\project", ["age", "species"], ["astronauts"]],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_recursive_double(self):
        expression = (
            r"\project_{species} \project_{age} \project_{name} "
            r"astronauts;"
        )
        expected = [
            [
                "\\project",
                ["species"],
                ["\\project", ["age"], ["\\project", ["name"], ["astronauts"]]],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_basic(self):
        expression = r"\project;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_relation(self):
        expression = r"\project_{species};"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_attributes(self):
        expression = r"\project astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_empty_attribute_list(self):
        expression = r"\project_{} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_space_before_attributes(self):
        expression = r"\project _{species} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_attribute_not_delim(self):
        expression = r"\project _{species age} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )



from rapt2.treebrd.attributes import AttributeList
from rapt2.treebrd.condition_node import (BinaryConditionalOperator,
                                          BinaryConditionNode,
                                          IdentityConditionNode)
from rapt2.treebrd.errors import (AttributeReferenceError, InputError,
                                  InputReferenceError)
from rapt2.treebrd.node import (AssignNode, Node, Operator, ProjectNode,
                                RenameNode, SelectNode, UnaryNode)
from tests.treebrd.test_node import NodeTestCase


class UnaryTestCase(NodeTestCase):
    pass


class TestUnaryNode(UnaryTestCase):
    def test_operator_from_init(self):
        actual = UnaryNode(Operator.select, self.alpha).operator
        self.assertEqual(Operator.select, actual)

    def test_name_when_init_has_none_is_child_name(self):
        node = UnaryNode(Operator.relation, self.alpha)
        expected = "alpha"
        self.assertEqual(expected, node.name)

    def test_name_when_init_has_name(self):
        node = UnaryNode(Operator.relation, self.alpha, "a_name")
        self.assertEqual("a_name", node.name)

    def test_child_after_init(self):
        node = UnaryNode(Operator.relation, self.alpha)
        self.assertEqual(self.alpha, node.child)

    def test_attributes_are_copied_from_child_when_init(self):
        expected = AttributeList(self.schema.get_attributes("beta"), "beta").to_list()
        node = UnaryNode(Operator.relation, self.beta, self.schema)
        self.assertEqual(expected, node.attributes.to_list())


class TestUnaryNodeEquality(UnaryTestCase):
    def test_equality_when_identical(self):
        node = UnaryNode(Operator.relation, Node(Operator.relation))
        same = node
        self.assertTrue(node == same)

    def test_equality_when_same_operator_and_name_and_child(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        twin = UnaryNode(Operator.relation, self.alpha, "borg")
        self.assertTrue(node == twin)

    def test_non_equality_when_different_operator(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.project, self.alpha, "borg")
        self.assertTrue(node != other)

    def test_non_equality_when_different_name(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.relation, self.alpha, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_different_child(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.relation, self.beta, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_different_operator_and_name(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.project, self.alpha, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_different_operator_and_child(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.project, self.beta, "borg")
        self.assertTrue(node != other)

    def test_non_equality_when_different_name_and_child(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.relation, self.beta, "other")
        self.assertTrue(node != other)

    def test_non_equality_when_different_operator_and_name_and_child(self):
        node = UnaryNode(Operator.relation, self.alpha, "borg")
        other = UnaryNode(Operator.project, self.beta, "other")
        self.assertTrue(node != other)


class TestSelect(UnaryTestCase):
    def test_operator_is_correct(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("42"),
        )
        actual = SelectNode(self.alpha, condition).operator
        self.assertEqual(Operator.select, actual)

    def test_condition_when_init_has_condition(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("42"),
        )
        actual = SelectNode(self.alpha, condition).conditions
        self.assertEqual(condition, actual)

    def test_condition_when_multiple_conditions(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("a1"),
                IdentityConditionNode("43"),
            ),
        )
        node = SelectNode(self.alpha, condition)
        self.assertEqual(condition, node.conditions)

    def test_condition_when_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.GREATER_THAN,
            IdentityConditionNode("alpha.a1"),
            IdentityConditionNode("41"),
        )
        node = SelectNode(self.alpha, condition)
        self.assertEqual(condition, node.conditions)

    def test_condition_when_multiple_conditions_with_prefix(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.AND,
            BinaryConditionNode(
                BinaryConditionalOperator.GREATER_THAN,
                IdentityConditionNode("alpha.a1"),
                IdentityConditionNode("41"),
            ),
            BinaryConditionNode(
                BinaryConditionalOperator.LESS_THAN,
                IdentityConditionNode("alpha.a1"),
                IdentityConditionNode("43"),
            ),
        )
        node = SelectNode(self.alpha, condition)
        self.assertEqual(condition, node.conditions)

    def test_exception_when_first_attribute_in_condition_is_wrong(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a2"),
            IdentityConditionNode("42"),
        )
        self.assertRaises(AttributeReferenceError, SelectNode, self.alpha, condition)

    def test_exception_when_second_attribute_in_condition_is_wrong(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a1"),
            IdentityConditionNode("a2"),
        )
        self.assertRaises(AttributeReferenceError, SelectNode, self.alpha, condition)

    def test_exception_when_both_attributes_in_condition_are_wrong(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("a2"),
            IdentityConditionNode("a3"),
        )
        self.assertRaises(AttributeReferenceError, SelectNode, self.alpha, condition)

    def test_exception_when_prefix_in_condition_is_wrong(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.EQUAL,
            IdentityConditionNode("beta.a1"),
            IdentityConditionNode("42"),
        )
        self.assertRaises(AttributeReferenceError, SelectNode, self.alpha, condition)

    def test_exception_when_ambiguous_attributes(self):
        condition = BinaryConditionNode(
            BinaryConditionalOperator.LESS_THAN,
            IdentityConditionNode("d1"),
            IdentityConditionNode("5"),
        )
        self.assertRaises(
            AttributeReferenceError, SelectNode, self.ambiguous, condition
        )


class TestProject(UnaryTestCase):
    def test_operator_is_correct(self):
        actual = ProjectNode(self.alpha, []).operator
        self.assertEqual(Operator.project, actual)

    def test_attributes_when_init_with_empty_list(self):
        node = ProjectNode(self.beta, [])
        self.assertEqual([], node.attributes.to_list())

    def test_attributes_when_init_with_all_attributes(self):
        expected = AttributeList(self.schema.get_attributes("beta"), "beta").to_list()
        node = ProjectNode(self.beta, ["b1", "b2"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_init_with_some_attributes(self):
        expected = ["gamma.c2", "gamma.c3"]
        node = ProjectNode(self.gamma, ["c2", "c3"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_init_with_some_attributes_different_order(self):
        expected = ["gamma.c3", "gamma.c2"]
        node = ProjectNode(self.gamma, ["c3", "c2"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_init_with_single_attribute(self):
        expected = ["gamma.c2"]
        node = ProjectNode(self.gamma, ["c2"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_init_with_project_child(self):
        expected = ["gamma.c1"]
        child = ProjectNode(self.gamma, ["c1", "c2"])
        node = ProjectNode(child, ["c1"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_prefixed(self):
        expected = ["gamma.c1", "gamma.c2"]
        node = ProjectNode(self.gamma, ["gamma.c1", "gamma.c2"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_attributes_when_mixed(self):
        expected = ["gamma.c1", "gamma.c2"]
        node = ProjectNode(self.gamma, ["gamma.c1", "c2"])
        self.assertEqual(expected, node.attributes.to_list())

    def test_exception_when_attribute_does_not_exist(self):
        self.assertRaises(AttributeReferenceError, ProjectNode, self.alpha, ["void"])

    def test_exception_when_attribute_no_longer_exists(self):
        child = ProjectNode(self.gamma, ["c1", "c2"])
        self.assertRaises(AttributeReferenceError, ProjectNode, child, ["c3"])

    def test_exception_when_prefix_is_wrong(self):
        self.assertRaises(InputReferenceError, ProjectNode, self.beta, ["alpha.c2"])

    def test_exception_when_ambiguous_attributes(self):
        self.assertRaises(AttributeReferenceError, ProjectNode, self.ambiguous, "d1")


class TestRename(TestUnaryNode):
    def test_operator_is_correct(self):
        actual = RenameNode(self.alpha, "other", [], self.schema).operator
        self.assertEqual(Operator.rename, actual)

    def test_name_when_renamed(self):
        actual = RenameNode(self.alpha, "mask", [], self.schema).name
        self.assertEqual("mask", actual)

    def test_attributes_when_renamed(self):
        expected = ["beta.a", "beta.b"]
        node = RenameNode(self.beta, None, ["a", "b"], self.schema)
        self.assertEqual(expected, node.attributes.to_list())

    def test_name_when_name_is_not_renamed(self):
        node = RenameNode(self.beta, None, ["a", "b"], self.schema)
        self.assertEqual("beta", node.name)

    def test_name_when_name_and_attributes_renamed(self):
        node = RenameNode(self.beta, "mask", ["a", "b"], self.schema)
        self.assertEqual("mask", node.name)

    def test_attributes_when_name_and_attributes_renamed(self):
        expected = ["mask.a", "mask.b"]
        node = RenameNode(self.beta, "mask", ["a", "b"], self.schema)
        self.assertEqual(expected, node.attributes.to_list())

    def test_no_exception_when_renaming_ambiguous_relation(self):
        expected = ["mask.a", "mask.b"]
        node = RenameNode(self.ambiguous, "mask", ["a", "b"], self.schema)
        self.assertEqual(expected, node.attributes.to_list())

    def test_exception_when_too_few_attributes(self):
        self.assertRaises(InputError, RenameNode, self.beta, None, ["a"], self.schema)

    def test_exception_when_too_many_attributes(self):
        self.assertRaises(
            InputError, RenameNode, self.beta, None, ["a", "b", "c"], self.schema
        )

    def test_exception_when_name_conflicts(self):
        self.assertRaises(
            InputError, RenameNode, self.beta, "alpha", ["a", "b"], self.schema
        )

    def test_when_ambiguous_attributes(self):
        self.assertRaises(
            InputError, RenameNode, self.beta, "beta", ["b1", "b1"], self.schema
        )


class TestAssignment(TestUnaryNode):
    def test_operator_is_correct(self):
        actual = AssignNode(self.alpha, "omega", [], self.schema).operator
        self.assertEqual(Operator.assign, actual)

    def test_name_when_given(self):
        actual = AssignNode(self.alpha, "omega", [], self.schema).name
        self.assertEqual("omega", actual)

    def test_attributes_when_renamed(self):
        node = AssignNode(self.beta, "omega", ["a", "b"], self.schema)
        expected = ["omega.a", "omega.b"]
        self.assertEqual(expected, node.attributes.to_list())

    def test_ambiguous_attributes_in_child(self):
        node = AssignNode(self.ambiguous, "omega", ["a", "b"], self.schema)
        expected = ["omega.a", "omega.b"]
        self.assertEqual(expected, node.attributes.to_list())

    def test_schema_is_updated_after_assign_with_no_explicit_attributes(self):
        AssignNode(self.beta, "omega", [], self.schema)
        self.assertTrue(self.schema.contains("omega"))
        self.assertEqual(
            self.schema.get_attributes("beta"), self.schema.get_attributes("omega")
        )

    def test_schema_is_updated_after_assign_with_attributes(self):
        AssignNode(self.gamma, "omega", ["a", "b", "c"], self.schema)
        self.assertTrue(self.schema.contains("omega"))
        self.assertEqual(["a", "b", "c"], self.schema.get_attributes("omega"))

    def test_exception_when_missing_name(self):
        self.assertRaises(InputError, AssignNode, self.alpha, None, ["a"], self.schema)

    def test_exception_when_too_few_attributes(self):
        self.assertRaises(
            InputError, AssignNode, self.beta, "omega", ["a"], self.schema
        )

    def test_exception_when_too_many_attributes(self):
        self.assertRaises(
            InputError, AssignNode, self.beta, "omega", ["a", "b", "c"], self.schema
        )

    def test_when_ambiguous_attributes(self):
        self.assertRaises(
            InputError, AssignNode, self.beta, "omega", ["o1", "o1"], self.schema
        )

    def test_exception_when_name_conflicts(self):
        self.assertRaises(
            InputError, RenameNode, self.beta, "alpha", ["a", "b"], self.schema
        )



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestAssignment(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_assign_relation(self):
        expression = "asts := astronauts;"
        expected = [[["asts"], ":=", ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_assign_expression(self):
        expression = "ast := \\project_{name} astronauts;"
        expected = [[["ast"], ":=", ["\\project", ["name"], ["astronauts"]]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_multiple_assign_expressions(self):
        expression = (
            "names := \\project_{name} astronauts;"
            "species := \\project_{species} astronauts;"
        )
        expected = [
            [["names"], ":=", ["\\project", ["name"], ["astronauts"]]],
            [["species"], ":=", ["\\project", ["species"], ["astronauts"]]],
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_assign_attributes(self):
        expression = "ast(age, happiness) := astronauts;"
        expected = [[["ast", ["age", "happiness"]], ":=", ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_assign_attributes_spacing(self):
        expression = "ast ( age , happiness ) := astronauts;"
        expected = [[["ast", ["age", "happiness"]], ":=", ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exceptions

    def test_no_expressions(self):
        expression = ":=;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_exp_no_right(self):
        expression = "A := ;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_no_left(self):
        expression = ":= A;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    def test_no_attributes(self):
        expression = "ast() := A;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression, CoreGrammar
        )

    # Definition tests

    def test_definition_with_assignments(self):
        expression = "R1(a, b); R2 := R1; R3(c, d);"
        expected = [
            ["r1", ["a", "b"]],
            [["r2"], ":=", ["r1"]],
            ["r3", ["c", "d"]],
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_only(self):
        expression = "R(a, b, c);"
        expected = [["r", ["a", "b", "c"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_single_attribute(self):
        expression = "R(a);"
        expected = [["r", ["a"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestDefinition(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_definition_single_attribute(self):
        expression = "R(a);"
        expected = [["r", ["a"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_multiple_attributes(self):
        expression = "R(a, b, c);"
        expected = [["r", ["a", "b", "c"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_two_attributes(self):
        expression = "R(a, b);"
        expected = [["r", ["a", "b"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_with_spacing(self):
        expression = "R ( a , b , c ) ;"
        expected = [["r", ["a", "b", "c"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_relation_name_only(self):
        # Test that definition requires attributes
        expression = "R();"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_no_attributes(self):
        # Test that definition requires at least one attribute
        # R; parses as a relation expression, not a definition
        expression = "R;"
        expected = [["r"]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_without_terminator(self):
        # Test that definition without terminator fails in statement context
        expression = "R(a, b, c)"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_parsing_only(self):
        # Test that definition parsing works without terminator
        expression = "R(a, b, c)"
        expected = [["r", ["a", "b", "c"]]]
        actual = self.parser.definition.parseString(expression, parseAll=True).asList()
        self.assertEqual(expected, actual)

    def test_multiple_definitions(self):
        expression = "R1(a, b); R2(c, d); R3(e);"
        expected = [
            ["r1", ["a", "b"]],
            ["r2", ["c", "d"]],
            ["r3", ["e"]],
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_definition_mixed_with_assignments(self):
        expression = "R1(a, b); R2 := R1; R3(c, d);"
        expected = [
            ["r1", ["a", "b"]],
            [["r2"], ":=", ["r1"]],
            ["r3", ["c", "d"]],
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exception tests

    def test_definition_missing_left_paren(self):
        expression = "R a, b, c);"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_missing_right_paren(self):
        expression = "R(a, b, c;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_empty_attributes(self):
        expression = "R();"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_trailing_comma(self):
        expression = "R(a, b, c,);"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_leading_comma(self):
        expression = "R(, a, b, c);"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_definition_double_comma(self):
        expression = "R(a, , b, c);"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )


import unittest

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestJoin(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_single(self):
        expression = "astronauts \\join sputnik1_crew;"
        expected = [[["astronauts"], "\\join", ["sputnik1_crew"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_double(self):
        expression = "astronauts \\join sputnik1_crew \\join sputnik2_crew;"
        expected = [
            [["astronauts"], "\\join", ["sputnik1_crew"], "\\join", ["sputnik2_crew"]]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_triple(self):
        expression = (
            "astronauts \\join sputnik1_crew \\join sputnik2_crew \\join sputnik3_crew;"
        )
        expected = [
            [
                ["astronauts"],
                "\\join",
                ["sputnik1_crew"],
                "\\join",
                ["sputnik2_crew"],
                "\\join",
                ["sputnik3_crew"],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_precedence_left(self):
        expression = (
            "astronauts \\join sputnik1_crew "
            "\\union sputnik2_crew \\union sputnik3_crew;"
        )
        expected = [
            [
                [["astronauts"], "\\join", ["sputnik1_crew"]],
                "\\union",
                ["sputnik2_crew"],
                "\\union",
                ["sputnik3_crew"],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_precedence_right(self):
        expression = (
            "astronauts \\union sputnik1_crew "
            "\\union "
            "sputnik2_crew \\join sputnik3_crew;"
        )
        expected = [
            [
                ["astronauts"],
                "\\union",
                ["sputnik1_crew"],
                "\\union",
                [["sputnik2_crew"], "\\join", ["sputnik3_crew"]],
            ]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)



from typing import TypeVar

from pyparsing import (CaselessKeyword, Group, Keyword, OneOrMore, Optional,
                       Suppress, alphanums, delimitedList, infixNotation,
                       opAssoc)

from .condition_grammar import ConditionGrammar
from .syntax import CoreSyntax

TCoreSyntax = TypeVar("TCoreSyntax", bound=CoreSyntax)


class CoreGrammar(ConditionGrammar[TCoreSyntax]):
    """
    A parser that recognizes a core relational algebra grammar.

    The rules are annotated with their BNF equivalents. For a complete
    specification refer to the associated grammar file.
    """

    syntax: TCoreSyntax

    def __init__(self, syntax: TCoreSyntax = CoreSyntax()):
        """
        Initializes a CoreGrammar. Uses the default syntax if none
        is provided.

        :param syntax: a syntax for this grammar.
        """
        super().__init__(syntax)

    def parse(self, instring):
        instring = r"" + instring
        return self.statements.parseString(instring, parseAll=True)

    @property
    def attribute_list(self):
        """
        attribute_list ::= attribute_name | attribute_list delim attribute_name
        """
        return delimitedList(self.attribute_name, delim=self.syntax.delim)

    @property
    def attribute_reference_list(self):
        """
        attribute_reference_list ::= attribute_reference |
                                 attribute_list delim attribute_reference
        """
        return delimitedList(self.attribute_reference, delim=self.syntax.delim)

    @property
    def relation(self):
        """
        relation_expr ::= relation_name
        """
        return Group(self.relation_name)

    # ## Unary Expressions ###

    @property
    def project(self):
        """
        project_expr ::= project param_start attribute_list param_stop
            expression
        """
        return self.parametrize(self.syntax.project_op, self.attribute_reference_list)

    @property
    def select(self):
        """
        select_expr ::= select param_start conditions param_stop expression
        """
        return self.parametrize(self.syntax.select_op, self.conditions)

    @property
    def rename(self):
        """
        rename_expr ::= rename param_start rename_parameters param_stop
            expression

        rename_parameters ::= relation_name |
            paren_left attribute_list paren_right |
            relation_name paren_left attribute_list paren_right
        """
        params = self.relation_name ^ (
            Optional(self.relation_name) + self.parenthesize(self.attribute_list)
        )
        return self.parametrize(self.syntax.rename_op, params)

    @property
    def unary_op(self):
        """
        rename_expr ::= rename param_start rename_parameters param_stop
            expression
        """
        return self.project | self.select | self.rename

    # Binary Operators (Precedence 1)

    @property
    def join(self):
        """
        join_expr ::= expression join expression
        """
        return CaselessKeyword(self.syntax.join_op)

    @property
    def binary_op_p1(self):
        """
        Binary operators with precedence of one.
        """
        return self.join

    # Binary Operators (Precedence 2)

    @property
    def union(self):
        """
        union_expr ::= expression union expression
        """
        return CaselessKeyword(self.syntax.union_op)

    @property
    def difference(self):
        """
        difference_expr ::= expression difference expression
        """
        return CaselessKeyword(self.syntax.difference_op)

    @property
    def binary_op_p2(self):
        """
        Binary operators with precedence of two.
        """
        return self.union | self.difference

    @property
    def expression(self):
        """
        A relation algebra expression. An expression is either a relation or a
        combination of relations that uses the previously defined operators
        and follows precedence rules.
        """
        return infixNotation(
            self.relation,
            [
                (self.unary_op, 1, opAssoc.RIGHT),
                (self.binary_op_p1, 2, opAssoc.LEFT),
                (self.binary_op_p2, 2, opAssoc.LEFT),
            ],
            lpar=self.syntax.paren_left,
            rpar=self.syntax.paren_right,
        )

    @property
    def assignment(self):
        """
        assignment ::= relation_name assign expression |
            relation_name paren_left attribute_list paren_right
            assign expression
        """
        lhs = Group(
            self.relation_name + Optional(self.parenthesize(self.attribute_list))
        )
        return Group(lhs + Keyword(self.syntax.assign_op) + self.expression)

    @property
    def definition(self):
        """
        definition ::= relation_name paren_left attribute_list paren_right
        """
        return Group(
            self.relation_name + self.parenthesize(self.attribute_list)
        )

    @property
    def statement(self):
        """
        A terminated relational algebra statement.
        """
        return (self.assignment ^ self.definition ^ self.expression) + Suppress(self.syntax.terminator)

    @property
    def statements(self):
        """
        An ordered collection of relational algebra statements.
        """
        return OneOrMore(self.statement)

    def parenthesize(self, parser):
        return (
            Suppress(self.syntax.paren_left)
            + Group(parser)
            + Suppress(self.syntax.paren_right)
        )

    def parameter(self, parser):
        """
        Return a parser the parses parameters.
        """
        return (
            Suppress(self.syntax.params_start).leaveWhitespace()
            + Group(parser)
            + Suppress(self.syntax.params_stop)
        )

    def parametrize(self, operator, params):
        """
        Return a parser that parses an operator with parameters.
        """
        return CaselessKeyword(operator, identChars=alphanums) + self.parameter(params)

    def is_unary(self, operator):
        return operator in {
            self.syntax.select_op,
            self.syntax.project_op,
            self.syntax.rename_op,
        }

    def is_binary(self, operator):
        return operator in {
            self.syntax.difference_op,
            self.syntax.union_op,
            self.syntax.join_op,
        }



import unittest

from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar


class TestProject(unittest.TestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_rename_relation(self):
        expression = "\\rename_{species} astronauts;"
        expected = [["\\rename", ["species"], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_rename_attribute(self):
        expression = "\\rename_{ (name)} astronauts;"
        expected = [["\\rename", [["name"]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_rename_multiple_attributes(self):
        expression = "\\rename_{(name, age, alive)} astronauts;"
        expected = [["\\rename", [["name", "age", "alive"]], ["astronauts"]]]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    def test_rename_relation_and_multiple_attributes(self):
        expression = "\\rename_{nautical (name, age, alive)} astronauts;"
        expected = [
            ["\\rename", ["nautical", ["name", "age", "alive"]], ["astronauts"]]
        ]
        actual = self.parser.statements.parseString(expression).asList()
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_basic(self):
        expression = "\\rename;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_exp_no_relation(self):
        expression = "\\rename_{species};"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_exp_no_attributes(self):
        expression = "\\rename astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_exp_empty_attribute_list(self):
        expression = "\\rename_{} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_exp_space_before_attributes(self):
        expression = "\\rename _{species} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )

    def test_exp_attribute_not_delim(self):
        expression = "\\rename _{species age} astronauts;"
        self.assertRaises(
            ParseException, self.parser.statements.parseString, expression
        )



import itertools
from collections import namedtuple

from .errors import AttributeReferenceError, InputError


class Attribute(namedtuple("Attribute", ["name", "prefix"])):
    """
    An Attribute is a relational algebra attribute. Attributes have optional
    prefixes which reference the relation they belong to.
    """

    @property
    def prefixed(self):
        if self.prefix:
            return "{pr}.{nm}".format(pr=self.prefix, nm=self.name)
        else:
            return self.name

    def __eq__(self, other):
        if type(self) is type(other):
            return self.prefixed == other.prefixed
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.prefixed)


class AttributeList:
    """
    A AttributeList is an ordered collection of relational algebra attributes.

    Attributes can have a prefix, which reference the relation they belong to.
    """

    @classmethod
    def merge(cls, first, second):
        """
        Return an AttributeList that is the result of merging first with second.
        """
        merged = AttributeList([], None)

        assert isinstance(first, AttributeList)
        assert isinstance(second, AttributeList)
        merged._contents = first._contents[:]
        merged._contents += second._contents[:]

        return merged

    @staticmethod
    def has_duplicates(collection):
        """
        Return True if the collection has duplicate elements.
        """
        return len(set(collection)) != len(collection)

    def __init__(self, names, prefix):
        self._contents = []
        self.extend(names, prefix)

    def __str__(self):
        """
        Return a comma delimitted string of prefixed attribute names.
        """
        return ", ".join(self.to_list())

    def __len__(self):
        return len(self._contents)

    def __eq__(self, other):
        if type(self) is type(other):
            return self.to_list() == other.to_list()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        return iter(self._contents)

    @property
    def contents(self):
        """
        Return a list of the Attributes in the AttributeList.
        """
        return self._contents

    @property
    def names(self):
        """
        Return a list of the names of the Attributes in the AttributeList.
        """
        return [name for name, _ in self._contents]

    def validate(self, references):
        """
        Check if all references exist and are unambiguous.
        """
        for reference in references:
            self.get_attribute(reference)

    def to_list(self):
        """
        Return a list of attributes. If an attribute has a prefix, return a
        prefixed attribute of the form: prefix.attribute.
        """
        return [attribute.prefixed for attribute in self._contents]

    def get_attribute(self, reference):
        """
        Return the attribute that matches the reference. Raise an error if
        the attribute cannot be found, or if there is more then one match.
        """
        prefix, _, name = reference.rpartition(".")

        match = None
        for attribute in self._contents:
            if name == attribute.name and (not prefix or prefix == attribute.prefix):
                if match:
                    raise AttributeReferenceError(
                        "Ambiguous attribute reference: {}.".format(attribute.name)
                    )
                else:
                    match = attribute

        if match:
            return match
        raise AttributeReferenceError("Attribute does not exist: {}".format(reference))

    def extend(self, attributes, prefix):
        """
        Add the attributes with the specified prefix to the end of the attribute
        list.
        """

        self._contents += [Attribute(attr, prefix) for attr in attributes]

    def trim(self, restriction_list):
        """
        Trim and reorder the attributes to the specifications in a restriction
        list.
        """
        replacement = []
        for reference in restriction_list:
            replacement.append(self.get_attribute(reference))

        if self.has_duplicates(replacement):
            raise AttributeReferenceError("Duplicate attribute reference.")
        self._contents = replacement

    def rename(self, names, prefix):
        """
        Rename the Attributes' names, prefixes, or both. If names or prefix
        evaluates to None, the old version is used.
        Resulting names must be unambiguous.
        :param names: A list of new names for each attribute or an empty list.
        :param prefix: A new prefix for the name or None
        """
        if names:
            if len(names) != len(self._contents):
                raise InputError("Attribute count mismatch.")
            if self.has_duplicates(names):
                raise InputError("Attributes are ambiguous.")
        else:
            # If the attributes are not renamed, but the relation / prefix is,
            # there is a risk of creating two or more attributes with the
            # same name and prefix.
            if prefix and self.has_duplicates(self.names):
                raise AttributeReferenceError("Attributes are ambiguous.")

        replacement = []
        for old, name in itertools.zip_longest(self._contents, names):
            new_name = name or old.name
            new_prefix = prefix or old.prefix
            replacement.append(Attribute(new_name, new_prefix))
        self._contents = replacement



class TreeBRDError(Exception):
    """
    Base class for exceptions in this module.
    """

    pass


class InputError(TreeBRDError):
    pass


class InputReferenceError(InputError):
    pass


class RelationReferenceError(InputReferenceError):
    pass


class AttributeReferenceError(InputReferenceError):
    pass



from typing import TypeVar

from pyparsing import CaselessKeyword, Group, oneOf, opAssoc
from pyparsing.helpers import infix_notation

from .proto_grammar import ProtoGrammar
from .syntax import ConditionSyntax

TConditionSyntax = TypeVar("TConditionSyntax", bound=ConditionSyntax)


def get_attribute_references(instring):
    """
    Return a list of attribute references in the condition expression.

    attribute_reference ::= relation_name "." attribute_name | attribute_name

    :param instring: a condition expression.
    :return: a list of attribute references.
    """
    parsed = ConditionGrammar().conditions.parseString(instring)
    result = parsed if isinstance(parsed[0], str) else parsed[0]

    attribute_refs = []

    def collect_attributes(item):
        if hasattr(item, "attribute_reference") and item.attribute_reference:
            attribute_refs.extend(item.attribute_reference.asList())
        if hasattr(item, "__iter__") and not isinstance(item, str):
            for subitem in item:
                collect_attributes(subitem)

    collect_attributes(result)
    return attribute_refs


class ConditionGrammar(ProtoGrammar[TConditionSyntax]):
    """
    A grammar for condition expressions.
    """

    def __init__(self, syntax: TConditionSyntax = ConditionSyntax()):
        """
        Initializes a ConditionGrammar. Uses the default syntax if none
        is provided.

        :param syntax: a syntax for this grammar.
        """
        super().__init__(syntax)

    @property
    def comparator_op(self):
        return oneOf(
            [
                self.syntax.equal_op,
                self.syntax.not_equal_op,
                self.syntax.not_equal_alt_op,
                self.syntax.less_than_op,
                self.syntax.less_than_equal_op,
                self.syntax.greater_than_op,
                self.syntax.greater_than_equal_op,
            ]
        )

    @property
    def not_op(self):
        return CaselessKeyword(self.syntax.not_op)

    @property
    def logical_binary_op(self):
        """
        logical_binary_op ::=  and_op | or_op
        """
        return CaselessKeyword(self.syntax.and_op) | CaselessKeyword(self.syntax.or_op)

    @property
    def operand(self):
        """
        operand ::= identifier | string_literal | number
        """
        return (
            self.attribute_reference("attribute_reference*")
            | self.string_literal
            | self.number
        )

    @property
    def condition(self):
        """
        condition ::= operand comparator_op operand | not_op condition |
                  paren_left condition paren_right
        not_op and grouping rules are defined using infixNotation in
        conditions.
        """
        return Group(self.operand + self.comparator_op + self.operand)

    @property
    def conditions(self):
        """
        conditions ::= condition | condition logical_binary_op conditions
        Note: By default lpar and rpar arguments are suppressed.
        """
        return infix_notation(
            base_expr=self.condition,
            op_list=[
                (self.not_op, 1, opAssoc.RIGHT),
                (self.logical_binary_op, 2, opAssoc.RIGHT),
            ],
            lpar=self.syntax.paren_left,
            rpar=self.syntax.paren_right,
        )



# Relational algebra latex operators

PROJECT_OP = "\\pi"
RENAME_OP = "\\rho"
SELECT_OP = "\\sigma"
ASSIGN_OP = ":="

JOIN_OP = "\\times"
THETA_JOIN_OP = "\\times"
NATURAL_JOIN_OP = "\\bowtie"
FULL_OUTER_JOIN_OP = "\\fullouterjoin"
LEFT_OUTER_JOIN_OP = "\\leftouterjoin"
RIGHT_OUTER_JOIN_OP = "\\rightouterjoin"

DIFFERENCE_OP = "-"
UNION_OP = "\\cup"
INTERSECT_OP = "\\cap"

# Dependency operators
PRIMARY_KEY_OP = "\\text{PK}"
MULTIVALUED_DEPENDENCY_OP = "\\twoheadrightarrow"
FUNCTIONAL_DEPENDENCY_OP = "\\rightarrow"
INCLUSION_EQUIVALENCE_OP = "\\equiv"
INCLUSION_SUBSUMPTION_OP = "\\subseteq"

# Conditional operators
AND_OP = "\\land"
OR_OP = "\\lor"
EQUAL_OP = "\\eq"
NOT_EQUAL_OP = "\\neq"
LESS_THAN_OP = "\\lt"
LESS_THAN_EQUAL_OP = "\\leq"
GREATER_THAN_OP = "\\gt"
GREATER_THAN_EQUAL_OP = "\\geq"
NOT_OP = "\\neg"
DEFINED_OP = "\\text{defined}"



from typing import TypeVar

from pyparsing import Group, Literal, OneOrMore, Suppress

from .extended_grammar import ExtendedGrammar
from .syntax import DependencySyntax

TDependencySyntax = TypeVar("TDependencySyntax", bound=DependencySyntax)


class DependencyGrammar(ExtendedGrammar[TDependencySyntax]):
    """
    A parser that recognizes database dependency grammar rules.

    The rules are annotated with their BNF equivalents. For a complete
    specification refer to the associated grammar file.
    """

    syntax: TDependencySyntax

    def __init__(self, syntax: TDependencySyntax = DependencySyntax()):
        """
        Initializes a DependencyGrammar. Uses the default syntax if none
        is provided.

        :param syntax: a syntax for this grammar.
        """
        super().__init__(syntax)

    def parse(self, instring):
        instring = r"" + instring
        return self.statements.parseString(instring, parseAll=True)

    @property
    def pk(self):
        """
        pk ::= "pk"
        """
        return Literal(self.syntax.pk_op)

    @property
    def pk_dep(self):
        """
        pk_dep ::= pk param_start attribute_list param_stop relation_name
        """
        return Group(self.pk + self.parameter(self.attribute_list) + self.relation_name)

    @property
    def mvd(self):
        """
        mvd ::= "mvd"
        """
        return Literal(self.syntax.mvd_op)

    @property
    def select_or_relation(self):
        """
        select_or_relation ::= (select param_start conditions param_stop relation_name) | relation_name
        """
        return (
            (
                Literal(self.syntax.select_op)
                + self.parameter(self.conditions)
                + self.relation_name
            )
            | self.relation_name
        )

    @property
    def cond_dep_expr(self):
        """
        cond_dep_expr ::= param_start attribute_name delim attribute_name param_stop select_or_relation
        """
        return self.parameter(
            self.attribute_name + Suppress(self.syntax.delim) + self.attribute_name
        ) + self.select_or_relation

    @property
    def mvd_dep(self):
        """
        mvd_dep ::= mvd cond_dep_expr
        """
        return Group(self.mvd + self.cond_dep_expr)

    @property
    def fd(self):
        """
        fd ::= "fd"
        """
        return Literal(self.syntax.fd_op)

    @property
    def fd_dep(self):
        """
        fd_dep ::= fd cond_dep_expr
        """
        return Group(self.fd + self.cond_dep_expr)

    @property
    def inc_eq(self):
        """
        inc_eq ::= "inc="
        """
        return Literal(self.syntax.inc_equiv_op)

    @property
    def inc_expr(self):
        """
        inc_expr ::= param_start attribute_name delim attribute_name param_stop paren_left select_or_relation delim select_or_relation paren_right
        """
        return self.parameter(
            self.attribute_name + Suppress(self.syntax.delim) + self.attribute_name
        ) + self.parenthesize(
            Group(self.select_or_relation) + Suppress(self.syntax.delim) + Group(self.select_or_relation)
        )

    @property
    def inc_eq_dep(self):
        """
        inc_eq_dep ::= inc_eq inc_expr
        """
        return Group(self.inc_eq + self.inc_expr)

    @property
    def inc_subs(self):
        """
        inc_subs ::= "inc⊆"
        """
        return Literal(self.syntax.inc_subset_op)

    @property
    def inc_subs_dep(self):
        """
        inc_subs_dep ::= inc_subs inc_expr
        """
        return Group(self.inc_subs + self.inc_expr)

    @property
    def dep(self):
        """
        dep ::= pk_dep | mvd_dep | fd_dep | inc_eq_dep | inc_subs_dep
        """
        return (
            self.pk_dep
            | self.mvd_dep
            | self.fd_dep
            | self.inc_eq_dep
            | self.inc_subs_dep
        )

    @property
    def dep_statement(self):
        """
        dep_statement ::= dep terminate
        """
        return self.dep + Suppress(self.syntax.terminator)

    @property
    def statement(self):
        """
        statement ::= dep_statement | statement
        """
        return self.dep_statement | super().statement

    @property
    def statements(self):
        """
        An ordered collection of dependency and relational algebra statements.
        """
        return OneOrMore(self.statement)



from pyparsing import ParseException, alphas, nums

from rapt2.treebrd.grammars.proto_grammar import ProtoGrammar
from tests.treebrd.grammars.grammar_test_case import GrammarTestCase


class TestProtoGrammar(GrammarTestCase):
    def setUp(self):
        self.parser = ProtoGrammar()

    def test_character(self):
        expected = set(alphas + nums + "_")
        actual = set(self.parser.character)

        self.assertEqual(expected, actual)

    def test_number_integers(self):
        parse = self.parse_function(self.parser.number)

        self.assertEqual(parse("792"), ["792"])
        self.assertEqual(parse("0"), ["0"])
        self.assertEqual(parse("-42"), ["-42"])

    def test_number_floats(self):
        parse = self.parse_function(self.parser.number)

        self.assertEqual(parse("7.42"), ["7.42"])
        self.assertEqual(parse("0.0"), ["0.0"])
        self.assertEqual(parse(".576"), [".576"])
        self.assertEqual(parse("-42.7"), ["-42.7"])
        self.assertEqual(parse("-1042.789"), ["-1042.789"])

    def test_number_exp(self):
        parse = self.parse_function(self.parser.number)

        self.assertRaises(ParseException, parse, "_")
        self.assertRaises(ParseException, parse, "a")
        self.assertRaises(ParseException, parse, "--7")
        self.assertRaises(ParseException, parse, "++7")

    def test_number_floats_exp(self):
        parse = self.parse_function(self.parser.number)

        self.assertRaises(ParseException, parse, "7.a")
        self.assertRaises(ParseException, parse, "7,1")
        self.assertRaises(ParseException, parse, "3.7.4")
        self.assertRaises(ParseException, parse, ".")

    def test_identifier(self):
        parse = self.parse_function(self.parser.identifier)

        self.assertEqual(parse("y"), ["y"])
        self.assertEqual(parse("Q"), ["q"])
        self.assertEqual(parse("why"), ["why"])
        self.assertEqual(parse("why_not"), ["why_not"])
        self.assertEqual(parse("Life_Is_42"), ["life_is_42"])

    def test_identifier_exp(self):
        parse = self.parse_function(self.parser.identifier)

        self.assertRaises(ParseException, parse, "7_is_perfect")
        self.assertRaises(ParseException, parse, "_is_perfect")
        self.assertRaises(ParseException, parse, "1")
        self.assertRaises(ParseException, parse, "0")

    def test_literal(self):
        parse = self.parse_function(self.parser.string_literal)
        self.assertEqual(parse("'Single!'"), ["'Single!'"])
        self.assertEqual(parse('"Two is one."'), ["'Two is one.'"])
        self.assertEqual(parse("''"), ["''"])

    def test_literal_exp(self):
        parse = self.parse_function(self.parser.string_literal)

        self.assertRaises(ParseException, parse, "Do not quote me!")
        self.assertRaises(ParseException, parse, "Don't quote me!")
        self.assertRaises(ParseException, parse, "\"Mismatch'")



from pyparsing import ParseException

from rapt2.treebrd.grammars.core_grammar import CoreGrammar
from tests.treebrd.grammars.grammar_test_case import GrammarTestCase


class TestCoreGrammar(GrammarTestCase):
    def setUp(self):
        self.parser = CoreGrammar()

    def test_attribute_id(self):
        parse = self.parse_function(self.parser.attribute_reference)

        self.assertEqual(parse("attribute"), ["attribute"])
        self.assertEqual(parse("relation.attribute"), ["relation.attribute"])

    def test_attribute_id_exp(self):
        parse = self.parse_function(self.parser.attribute_reference)

        self.assertRaises(ParseException, parse, "relation.")
        self.assertRaises(ParseException, parse, ".attribute")
        self.assertRaises(ParseException, parse, "relation.attribute.attribute")
        self.assertRaises(ParseException, parse, ".")



from unittest import TestCase

from rapt2.treebrd.condition_node import (BinaryConditionalOperator,
                                          BinaryConditionNode,
                                          IdentityConditionNode,
                                          UnaryConditionalOperator,
                                          UnaryConditionNode)
from rapt2.treebrd.schema import Schema
from rapt2.transformers.qtree.qtree_translator import QTreeTranslator as QTreeTranslator
from rapt2.transformers.sql.sql_translator import SQLTranslator as SQLTranslator


class ConditionNodeTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def setUp(self):
        self.schema = Schema(
            {
                "alpha": ["a1"],
                "beta": ["b1", "b2"],
                "gamma": ["c1", "c2", "c3"],
            }
        )
        self.qtree_translator = QTreeTranslator()
        self.sql_translator = SQLTranslator()


class TestConditionNode(ConditionNodeTestCase):
    def test_identity_condition_node(self):
        node = IdentityConditionNode("alpha.a1")
        self.assertEqual(self.qtree_translator.translate_condition(node), "alpha.a1")
        self.assertEqual(self.sql_translator.translate_condition(node), "alpha.a1")
        self.assertEqual(node.attribute_references(), ["alpha.a1"])

    def test_unary_condition_node(self):
        child = IdentityConditionNode("alpha.a1")
        node = UnaryConditionNode(UnaryConditionalOperator.NOT, child)
        self.assertEqual(self.qtree_translator.translate_condition(node), "\\neg alpha.a1")
        self.assertEqual(self.sql_translator.translate_condition(node), "NOT alpha.a1")
        self.assertEqual(node.attribute_references(), ["alpha.a1"])

    def test_binary_condition_node(self):
        left = IdentityConditionNode("alpha.a1")
        right = IdentityConditionNode("beta.b1")
        node = BinaryConditionNode(BinaryConditionalOperator.AND, left, right)
        self.assertEqual(self.qtree_translator.translate_condition(node), "(alpha.a1 \\land beta.b1)")
        self.assertEqual(self.sql_translator.translate_condition(node), "(alpha.a1 AND beta.b1)")
        self.assertEqual(node.attribute_references(), ["alpha.a1", "beta.b1"])



from pyparsing import ParseException

from rapt2.treebrd.grammars.extended_grammar import ExtendedGrammar
from tests.treebrd.grammars.grammar_test_case import GrammarTestCase


class TestIntersect(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()
        self.parse = self.parse_function(self.parser.statements)

    def test_simple(self):
        expression = "zeppelin \\intersect floyd;"
        expected = [[["zeppelin"], "\\intersect", ["floyd"]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_multiple(self):
        expression = "zeppelin \\intersect floyd \\intersect doors;"
        expected = [[["zeppelin"], "\\intersect", ["floyd"], "\\intersect", ["doors"]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_intersect_precedence_when_before_set_operator(self):
        expression = "zeppelin \\intersect floyd \\union doors;"
        expected = [[[["zeppelin"], "\\intersect", ["floyd"]], "\\union", ["doors"]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_intersect_precedence_when_after_set_operator(self):
        expression = "zeppelin \\union floyd \\intersect doors;"
        expected = [[["zeppelin"], "\\union", [["floyd"], "\\intersect", ["doors"]]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_join(self):
        expression = "zeppelin \\intersect floyd \\join doors;"
        expected = [[["zeppelin"], "\\intersect", [["floyd"], "\\join", ["doors"]]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_unary(self):
        expression = "\\project_{albums} zeppelin \\intersect \\project_{albums} floyd;"
        expected = [
            [
                ["\\project", ["albums"], ["zeppelin"]],
                "\\intersect",
                ["\\project", ["albums"], ["floyd"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_no_relations(self):
        expression = "\\intersect;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_left_relation(self):
        expression = "roger \\intersect;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_right_relation(self):
        expression = "\\intersect waters;"
        self.assertRaises(ParseException, self.parse, expression)


class TestNaturalJoin(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()
        self.parse = self.parse_function(self.parser.statements)

    def test_simple(self):
        expression = "zeppelin \\natural_join floyd;"
        expected = [[["zeppelin"], "\\natural_join", ["floyd"]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_multiple(self):
        expression = "zeppelin \\natural_join floyd \\natural_join doors;"
        expected = [
            [["zeppelin"], "\\natural_join", ["floyd"], "\\natural_join", ["doors"]]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_same_precedence(self):
        expression = "zeppelin \\natural_join floyd \\join doors;"
        expected = [[["zeppelin"], "\\natural_join", ["floyd"], "\\join", ["doors"]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_precedence(self):
        expression = "zeppelin \\union floyd \\natural_join doors;"
        expected = [[["zeppelin"], "\\union", [["floyd"], "\\natural_join", ["doors"]]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_unary(self):
        expression = (
            "\\project_{albums} zeppelin \\natural_join \\project_{albums} floyd;"
        )
        expected = [
            [
                ["\\project", ["albums"], ["zeppelin"]],
                "\\natural_join",
                ["\\project", ["albums"], ["floyd"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_no_relations(self):
        expression = "\\natural_join;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_left_relation(self):
        expression = "roger \\natural_join;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_right_relation(self):
        expression = "\\natural_join waters;"
        self.assertRaises(ParseException, self.parse, expression)


class TestThetaJoin(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()
        self.parse = self.parse_function(self.parser.statements)

    def test_simple(self):
        expression = "zeppelin \\join_{year < 1975} floyd;"
        expected = [[["zeppelin"], "\\theta_join", [["year", "<", "1975"]], ["floyd"]]]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_multiple(self):
        expression = "zeppelin \\join_{year < 1975} floyd \\join_{year < 1960} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\theta_join",
                [["year", "<", "1975"]],
                ["floyd"],
                "\\theta_join",
                [["year", "<", "1960"]],
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_same_precedence(self):
        expression = "zeppelin \\join_{year < 1975} floyd \\join doors;"
        expected = [
            [
                ["zeppelin"],
                "\\theta_join",
                [["year", "<", "1975"]],
                ["floyd"],
                "\\join",
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_precedence(self):
        expression = "zeppelin \\union floyd \\join_{year < 1975} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\union",
                [["floyd"], "\\theta_join", [["year", "<", "1975"]], ["doors"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_unary(self):
        expression = (
            "\\project_{albums} zeppelin \\join_{year < 1975} \\project_{albums} floyd;"
        )
        expected = [
            [
                ["\\project", ["albums"], ["zeppelin"]],
                "\\theta_join",
                [["year", "<", "1975"]],
                ["\\project", ["albums"], ["floyd"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_basic(self):
        expression = "\\join_;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_no_relation_right(self):
        expression = "zeppelin \\join_{age < 42};"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_no_relation_left(self):
        expression = "\\join_{age < 42} floyd;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_empty_attribute_list(self):
        expression = "zeppelin \\join_{} floyd;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_space_before_attributes(self):
        expression = "zeppelin \\join _{age < 42} floyd;"
        self.assertRaises(ParseException, self.parse, expression)


class TestThreeValuedLogic(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()

    def test_defined_op(self):
        parse = self.parse_function(self.parser.defined_op)
        self.assertEqual(parse("defined"), ["defined"])
        self.assertEqual(parse("DEFINED"), ["defined"])

    def test_defined_condition(self):
        parse = self.parse_function(self.parser.defined_condition)
        self.assertEqual(parse("defined( apple )"), [["defined", "apple"]])

    def test_defined_condition_fragments(self):
        parse = self.parse_function(self.parser.defined_condition)
        self.assertRaises(ParseException, parse, "defined (")
        self.assertRaises(ParseException, parse, "defined ( )")

    def test_conditions_single_defined(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["defined", "apple"]]
        actual = parse("defined(apple)")
        self.assertEqual(actual, expected)

    def test_conditions_defined_with_parentheses(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["defined", "apple"]]
        actual = parse("(defined(apple))")
        self.assertEqual(actual, expected)

    def test_conditions_not_defined(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["not", ["defined", "apple"]]]
        actual = parse("not defined(apple)")
        self.assertEqual(actual, expected)

    def test_conditions_defined_and_defined(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["defined", "apple"], "and", ["defined", "pear"]]]
        actual = parse("defined(apple) and defined(pear)")
        self.assertEqual(actual, expected)

    def test_conditions_defined_and_defined_with_parentheses(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["defined", "apple"], "and", ["defined", "pear"]]]
        actual = parse("(defined(apple)) and (defined(pear))")
        self.assertEqual(actual, expected)

    def test_conditions_defined_or_defined(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["defined", "apple"], "or", ["defined", "pear"]]]
        actual = parse("(defined(apple)) or (defined(pear))")
        self.assertEqual(actual, expected)

    def test_conditions_unary_binary_defined_simple(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["not", ["defined", "apple"]], "and", ["defined", "pear"]]]
        actual = parse("not defined(apple) and defined(pear)")
        self.assertEqual(actual, expected)

    def test_conditions_unary_binary_defined_forced(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [["not", [["defined", "apple"], "and", ["defined", "pear"]]]]
        actual = parse("not (defined(apple) and defined(pear))")
        self.assertEqual(actual, expected)

    def test_conditions_binary_unary_defined_simple(self):
        parse = self.parse_function(self.parser.conditions)
        expected = [[["defined", "pear"], "or", ["not", ["defined", "apple"]]]]
        actual = parse("defined(pear) or not defined(apple)")
        self.assertEqual(expected, actual)


class TestFullOuterJoin(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()
        self.parse = self.parse_function(self.parser.statements)

    def test_simple(self):
        expression = "zeppelin \\full_outer_join_{year = year} floyd;"
        expected = [
            [["zeppelin"], "\\full_outer_join", [["year", "=", "year"]], ["floyd"]]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_multiple(self):
        expression = "zeppelin \\full_outer_join_{year = year} floyd \\full_outer_join_{year = year} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\full_outer_join",
                [["year", "=", "year"]],
                ["floyd"],
                "\\full_outer_join",
                [["year", "=", "year"]],
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_same_precedence(self):
        expression = "zeppelin \\full_outer_join_{year = year} floyd \\join doors;"
        expected = [
            [
                ["zeppelin"],
                "\\full_outer_join",
                [["year", "=", "year"]],
                ["floyd"],
                "\\join",
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_precedence(self):
        expression = "zeppelin \\union floyd \\full_outer_join_{year = year} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\union",
                [["floyd"], "\\full_outer_join", [["year", "=", "year"]], ["doors"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_unary(self):
        expression = "\\project_{albums} zeppelin \\full_outer_join_{year = year} \\project_{albums} floyd;"
        expected = [
            [
                ["\\project", ["albums"], ["zeppelin"]],
                "\\full_outer_join",
                [["year", "=", "year"]],
                ["\\project", ["albums"], ["floyd"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_no_relations(self):
        expression = "\\full_outer_join_;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_left_relation(self):
        expression = "roger \\full_outer_join_{year = year};"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_right_relation(self):
        expression = "\\full_outer_join_{year = year} floyd;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_empty_conditions(self):
        expression = "zeppelin \\full_outer_join_{} floyd;"
        self.assertRaises(ParseException, self.parse, expression)


class TestLeftOuterJoin(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()
        self.parse = self.parse_function(self.parser.statements)

    def test_simple(self):
        expression = "zeppelin \\left_outer_join_{year = year} floyd;"
        expected = [
            [["zeppelin"], "\\left_outer_join", [["year", "=", "year"]], ["floyd"]]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_multiple(self):
        expression = "zeppelin \\left_outer_join_{year = year} floyd \\left_outer_join_{year = year} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\left_outer_join",
                [["year", "=", "year"]],
                ["floyd"],
                "\\left_outer_join",
                [["year", "=", "year"]],
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_same_precedence(self):
        expression = "zeppelin \\left_outer_join_{year = year} floyd \\join doors;"
        expected = [
            [
                ["zeppelin"],
                "\\left_outer_join",
                [["year", "=", "year"]],
                ["floyd"],
                "\\join",
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_precedence(self):
        expression = "zeppelin \\union floyd \\left_outer_join_{year = year} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\union",
                [["floyd"], "\\left_outer_join", [["year", "=", "year"]], ["doors"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_unary(self):
        expression = "\\project_{albums} zeppelin \\left_outer_join_{year = year} \\project_{albums} floyd;"
        expected = [
            [
                ["\\project", ["albums"], ["zeppelin"]],
                "\\left_outer_join",
                [["year", "=", "year"]],
                ["\\project", ["albums"], ["floyd"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_no_relations(self):
        expression = "\\left_outer_join_;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_left_relation(self):
        expression = "roger \\left_outer_join_{year = year};"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_right_relation(self):
        expression = "\\left_outer_join_{year = year} floyd;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_empty_conditions(self):
        expression = "zeppelin \\left_outer_join_{} floyd;"
        self.assertRaises(ParseException, self.parse, expression)


class TestRightOuterJoin(GrammarTestCase):
    def setUp(self):
        self.parser = ExtendedGrammar()
        self.parse = self.parse_function(self.parser.statements)

    def test_simple(self):
        expression = "zeppelin \\right_outer_join_{year = year} floyd;"
        expected = [
            [["zeppelin"], "\\right_outer_join", [["year", "=", "year"]], ["floyd"]]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_multiple(self):
        expression = "zeppelin \\right_outer_join_{year = year} floyd \\right_outer_join_{year = year} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\right_outer_join",
                [["year", "=", "year"]],
                ["floyd"],
                "\\right_outer_join",
                [["year", "=", "year"]],
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_same_precedence(self):
        expression = "zeppelin \\right_outer_join_{year = year} floyd \\join doors;"
        expected = [
            [
                ["zeppelin"],
                "\\right_outer_join",
                [["year", "=", "year"]],
                ["floyd"],
                "\\join",
                ["doors"],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_precedence(self):
        expression = "zeppelin \\union floyd \\right_outer_join_{year = year} doors;"
        expected = [
            [
                ["zeppelin"],
                "\\union",
                [["floyd"], "\\right_outer_join", [["year", "=", "year"]], ["doors"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    def test_other_unary(self):
        expression = "\\project_{albums} zeppelin \\right_outer_join_{year = year} \\project_{albums} floyd;"
        expected = [
            [
                ["\\project", ["albums"], ["zeppelin"]],
                "\\right_outer_join",
                [["year", "=", "year"]],
                ["\\project", ["albums"], ["floyd"]],
            ]
        ]
        actual = self.parse(expression)
        self.assertEqual(expected, actual)

    # Exceptions

    def test_exp_no_relations(self):
        expression = "\\right_outer_join_;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_left_relation(self):
        expression = "roger \\right_outer_join_{year = year};"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_right_relation(self):
        expression = "\\right_outer_join_{year = year} floyd;"
        self.assertRaises(ParseException, self.parse, expression)

    def test_exp_empty_conditions(self):
        expression = "zeppelin \\right_outer_join_{} floyd;"
        self.assertRaises(ParseException, self.parse, expression)
